from pathlib import Path

PROMPT_DIR = Path(__file__).parent

def get_prompt(language: str) -> str:
    """Load prompt template for specified language."""
    prompt_path = PROMPT_DIR / f"prompt_gen_{language}.txt"
    if not prompt_path.exists():
        raise ValueError(f"No prompt template found for language: {language}")
    return prompt_path.read_text(encoding='utf-8')

__all__ = ['get_prompt']


from typing import List, Union
from pathlib import Path
from tqdm import tqdm
from ..models.qa_pair import Chunk
from ..converters.text import TextConverter
from ..strategies.chunk import ChunkStrategy

class ChunkGenerator:
    def __init__(
            self,
            text_converter: TextConverter,
            chunk_strategy: ChunkStrategy,
    ):
        self.text_converter = text_converter
        self.chunk_strategy = chunk_strategy
    
    def generate(self, input_path: Union[str, Path]) -> List[Chunk]:
        '''
        file/dir -> chunks
        handling both file or directory path
        '''
        chunks = []
        input_path = Path(input_path)

        files = [input_path] if input_path.is_file() else [path for path in input_path.glob('**/*') if path.is_file()]
        
        for file in tqdm(files, desc="Extract Content"):
            text = self.text_converter.convert(file)
            chunks.extend(self.chunk_strategy.split(source=file, text=text))
        
        return chunks








import json
import logging
from typing import Optional
from openai import AsyncOpenAI
from ..models.qa_pair import QAPair, Chunk

logger = logging.getLogger(__name__)

class QAGenerator:
    def __init__(self, client: AsyncOpenAI, llm_model: str, prompt_template: str):
        self.client = client
        self.llm_model = llm_model
        self.prompt_template = prompt_template
    '''
    Chunk -> QAPairs
    '''
    async def generate(self, chunk: Chunk) -> Optional[QAPair]:
        try:
            response = await self.client.completions.create(
                model=self.llm_model,
                prompt=self.prompt_template.format(text=chunk.content),
                max_tokens=1024,
            )
            
            response_text = response.choices[0].text
            if not self._is_valid_json(response_text):
                logger.info("Unable to receive the expected response in JSON format.")
                return None
            
            data = json.loads(response_text)
            return QAPair(
                instruction=data['instruction'],
                input=data['input'],
                output=data['output'],
                source=chunk,
            )
            
        except Exception as e:
            logger.error(f"Error during generating entry: {str(e)}")
            return None
    
    @staticmethod
    def _is_valid_json(response: str) -> bool:
        try:
            data = json.loads(response)
            return all(key in data for key in ['instruction', 'input', 'output'])
        except:
            return False


## AlpacaGen
import logging
from typing import Union
from pathlib import Path
from abc import ABC, abstractmethod
from markitdown import MarkItDown

class TextConverter(ABC):
    @abstractmethod
    def convert(self, input_path: Union[str, Path]) -> str:
        pass

class MarkItDownConverter(TextConverter):
    """ use MakeItDown as a converter """
    def convert(self, input_path: Union[str, Path]) -> str:
        """ input path should be a file, not a directory """
        md = MarkItDown()

        if isinstance(input_path, Path):
            input_path = str(input_path)

        try:
            extract_text = md.convert(input_path).text_content
            if not extract_text:
                logging.warning(f"{input_path} is empty after conversion.")
            
            return extract_text
        
        except Exception as e:
            logging.error(f"Error extracting file{input_path}: {str(e)}")
            raise


## AlpacaGen
from typing import Dict
from dataclasses import dataclass

@dataclass
class Chunk:
    content: str
    source: str
    idx: str  ## format: 01/17

@dataclass
class QAPair:
    instruction: str
    input: str
    output: str
    source: Chunk

    def to_dict(self) -> Dict[str, str]:
        return {
            'instruction': str(self.instruction),
            'input': str(self.input),
            'output': str(self.output),
            'source': repr(Chunk)
        }


from typing import List, Union
from pathlib import Path
from abc import ABC, abstractmethod
from ..models.qa_pair import Chunk
class ChunkStrategy(ABC):
    @abstractmethod
    def _split_text(self, text: str) -> List[str]:
        # customize in each strategy
        pass

    def split(self, source: Union[str, Path], text: str) -> List[Chunk]:
        chunks = self._split_text(text)
        return [
            Chunk(
                content=chunk,
                source=source,
                idx=f"{str(n+1).zfill(len(str(len(chunks))))}/{len(chunks)}"
            ) 
            for n, chunk in enumerate(chunks)
        ]

class RecursiveChunkStrategy(ChunkStrategy):
    def __init__(self, chunk_size: int = 4000, chunk_overlap: int = 200):
        from langchain_text_splitters import RecursiveCharacterTextSplitter
        self.splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap
        )
    
    def _split_text(self, text: str) -> List[str]:
        return self.splitter.split_text(text)


import asyncio
from typing import List, Tuple, Union
from pathlib import Path
from tqdm import tqdm
from tqdm.asyncio import tqdm_asyncio
from ..models.qa_pair import QAPair, Chunk
from ..converters.text import TextConverter
from ..strategies.chunk import ChunkStrategy
from .qa import QAGenerator
from .chunk import ChunkGenerator

class QADatasetGenerator:
    def __init__(
        self,
        qa_generator: QAGenerator,
        entries_per_chunk: int = 3
    ):
        self.qa_generator = qa_generator
        self.entries_per_chunk = entries_per_chunk

    async def generate(self, chunks: List[Chunk]) -> List[QAPair]:
        BATCH_SIZE = 20
        tasks = []
        dataset = []
        #TODO: modify input to Chunk, making QApairs with source info?
        for chunk in chunks:
            tasks.extend([self.qa_generator.generate(chunk) for _ in range(self.entries_per_chunk)])

        if not tasks:
            return chunks, []

        itrs = (len(tasks) // BATCH_SIZE) + (len(tasks) % BATCH_SIZE != 0)
        print(f"Total QAPairs: {len(tasks)}")
        
        for n, ith in enumerate(range(0, len(tasks), BATCH_SIZE)):
            batch_tasks = tasks[ith: ith + BATCH_SIZE]
            batch_dataset = await tqdm_asyncio.gather(
                *batch_tasks, 
                desc=f"Generate QA(Batch{n+1}/{itrs})"
            )
            dataset.extend(batch_dataset)
            
            del batch_tasks
            del batch_dataset
            await asyncio.sleep(0.5)
        
        dataset = [result for result in dataset if result is not None]
        return dataset
    
    # async def process_file(self, input_path: Union[str, Path]) -> Tuple[List[Chunk], List[QAPair]]:
    #     text = self.text_converter.convert(input_path)
    #     chunks: List[Chunk] = self.chunk_strategy.split(source=input_path, text=text)
        
    #     tasks = []
    #     for chunk in chunks:
    #         tasks.extend([self.qa_generator.generate(chunk.content) 
    #                      for _ in range(self.entries_per_chunk)])
        
    #     results = await tqdm_asyncio.gather(*tasks, desc="Generate QA")
    #     dataset: List[QAPair] = [result for result in results if result is not None]
        
    #     return chunks, dataset
    
    # async def process_directory(self, input_dir: Union[str, Path]) -> Tuple[List[Chunk], List[QAPair]]:
    #     BATCH_SIZE = 20
    #     input_dir = Path(input_dir)
    #     chunks = []
    #     tasks = []
    #     dataset = []
        
    #     paths = list(input_dir.glob('**/*'))
    #     files = [path for path in paths if path.is_file()]

    #     for input_path in tqdm(files, desc="Extract Content"):
    #         if input_path.is_file():
    #             text = self.text_converter.convert(input_path)
    #             chunks.extend(self.chunk_strategy.split(source=input_path, text=text))
        
    #     for chunk in chunks:
    #         tasks.extend([self.qa_generator.generate(chunk.content) 
    #                      for _ in range(self.entries_per_chunk)])

    #     if not tasks:
    #         return chunks, []

    #     itrs = (len(tasks) // BATCH_SIZE) + (len(tasks) % BATCH_SIZE != 0)
    #     print(f"Total QAPairs: {len(tasks)}")
        
    #     for n, ith in enumerate(range(0, len(tasks), BATCH_SIZE)):
    #         batch_tasks = tasks[ith: ith + BATCH_SIZE]
    #         batch_dataset = await tqdm_asyncio.gather(
    #             *batch_tasks, 
    #             desc=f"Generate QA(Batch{n+1}/{itrs})"
    #         )
    #         dataset.extend(batch_dataset)
            
    #         del batch_tasks
    #         del batch_dataset
    #         await asyncio.sleep(0.5)
        
    #     dataset = [result for result in dataset if result is not None]
    #     return chunks, dataset


from .main import AlpacaGen
from .models.qa_pair import QAPair, Chunk
from .strategies.chunk import ChunkStrategy, RecursiveChunkStrategy
from .converters.text import TextConverter, MarkItDownConverter

__version__ = "0.1.1"

__all__ = [
    'AlpacaGen',
    'QAPair',
    'Chunk',
    'ChunkStrategy',
    'RecursiveChunkStrategy',
    'TextConverter',
    'MarkItDownConverter',
]


import json
import logging
import asyncio
from typing import List, Optional, Union, Tuple, Any
from pathlib import Path
import nest_asyncio
from openai import AsyncOpenAI

from .converters.text import MarkItDownConverter
from .models.qa_pair import QAPair, Chunk
from .strategies.chunk import RecursiveChunkStrategy
from .generators.chunk import ChunkGenerator
from .generators.qa import QAGenerator
from .generators.dataset import QADatasetGenerator

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

LANGUAGES = ['zhtw', 'en']
LLM_CLIENTS = ['azure', 'openai']
DEFAULT_MODEL_DICT = {
    'azure': 'azure-gpt-4o',
    'openai': 'gpt-4o',
}
DEFAULT_PROMPT_PATHS = {
    lang: Path(__file__).parent / "prompt" / f"prompt_gen_{lang}.txt" 
    for lang in LANGUAGES
}

class AlpacaGen:
    def __init__(
            self,
            llm_client: str = None,
            llm_model: str = None,
            api_key: str = None,
            base_url: str = None,
    ):
        assert llm_client in LLM_CLIENTS, f"Specify your llm client, client should be one of {LLM_CLIENTS}"
        
        self.llm_client = llm_client
        self.llm_model = llm_model if llm_model else DEFAULT_MODEL_DICT[self.llm_client]
        self.api_key = api_key
        self.base_url = base_url
        self._client = None
        logging.basicConfig(level=logging.ERROR)

    async def _get_client(self) -> AsyncOpenAI:
        """Get or create AsyncOpenAI client."""
        if self._client is None:
            self._client = AsyncOpenAI(
                api_key=self.api_key,
                base_url=self.base_url,
            )
        return self._client

    def get_chunks(
        self,
        input_path: Union[str, Path],
        chunk_size: int = 4096,
    ) -> List[Chunk]:
        """Generate chunks from input file."""
        input_path = Path(input_path)
        if not input_path.exists():
            raise FileNotFoundError(f"Cannot find: {input_path}")

        chunk_generator = ChunkGenerator(
            text_converter=MarkItDownConverter(),
            chunk_strategy=RecursiveChunkStrategy(
                chunk_size=chunk_size,
                chunk_overlap=200,
            )
        )

        return chunk_generator.generate(input_path)

    async def _get_dataset_async(
        self,
        chunks: List[Chunk],
        language: str = 'zhtw',
        gen_prompt_path: Optional[Path] = None,
        entries_per_chunk: int = 3,
        output_path: Optional[Union[str, Path]] = None,
    ) -> List[QAPair]:
        """Internal async method for dataset generation."""
        assert language in LANGUAGES, f"Language should be one of {LANGUAGES}"

        def get_default_prompt(language: str) -> str:
            with DEFAULT_PROMPT_PATHS[language].open('r', encoding='utf-8') as f:
                return f.read()

        gen_prompt = (
            Path(gen_prompt_path).read_text(encoding='utf-8')
            if gen_prompt_path
            else get_default_prompt(language)
        )

        client = await self._get_client()
        
        dataset_generator = QADatasetGenerator(
            qa_generator=QAGenerator(
                client,
                self.llm_model,
                gen_prompt
            ),
            entries_per_chunk=entries_per_chunk
        )

        dataset = await dataset_generator.generate(chunks)

        if output_path:
            self.save_to_jsonl(dataset, output_path)

        return dataset

    def get_datasets(
        self,
        chunks: List[Chunk],
        language: str = 'zhtw',
        gen_prompt_path: Optional[Path] = None,
        entries_per_chunk: int = 3,
        output_path: Optional[Union[str, Path]] = None,
    ) -> List[QAPair]:
        """
        User-friendly synchronous method to generate datasets from chunks.
        
        Args:
            chunks: List of text chunks to process
            language: Language for prompt generation ('zhtw' or 'en')
            gen_prompt_path: Optional custom prompt file path
            entries_per_chunk: Number of QA pairs to generate per chunk
            output_path: Optional path to save the dataset
            
        Returns:
            List of QAPair objects representing the generated dataset
        """
        try:
            loop = asyncio.get_event_loop()
            if loop.is_running():
                nest_asyncio.apply()
            return loop.run_until_complete(
                self._get_dataset_async(
                    chunks,
                    language,
                    gen_prompt_path,
                    entries_per_chunk,
                    output_path
                )
            )
        except RuntimeError:
            return asyncio.run(
                self._get_dataset_async(
                    chunks,
                    language,
                    gen_prompt_path,
                    entries_per_chunk,
                    output_path
                )
            )

    def save_to_jsonl(self, dataset: List[QAPair], output_path: Union[str, Path]):
        """Save dataset (list of QAPair) to a JSONL file."""
        try:
            with Path(output_path).open('w', encoding='utf-8') as f:
                for qa_pair in dataset:
                    json_line = json.dumps(qa_pair.to_dict(), ensure_ascii=False)
                    f.write(json_line + '\n')
        except IOError as e:
            logger.error(f"Error writing to file {output_path}: {e}")
            raise
        except Exception as e:
            logger.error(f"Unexpected error occurred: {e}")
            raise


import pytest
from pathlib import Path
from alpacagen import (
    AlpacaGen,
    QAPair,
    Chunk,
    RecursiveChunkStrategy,
    MarkItDownConverter
)

# Test data
SAMPLE_TEXT = """# Test Document
This is a test document with multiple paragraphs.

## Section 1
This is the first section of the document.
It contains some text that should be processed.

## Section 2
This is the second section with more content.
"""

@pytest.fixture
def sample_text_file(tmp_path):
    file_path = tmp_path / "test_doc.txt"
    file_path.write_text(SAMPLE_TEXT)
    return file_path

@pytest.fixture
def mock_openai_client(mocker):
    mock_client = mocker.Mock()
    mock_response = mocker.Mock()
    mock_response.choices = [
        mocker.Mock(
            text='{"instruction": "Summarize the content", "input": "Given text", "output": "Summary"}'
        )
    ]
    mock_client.completions.create.return_value = mock_response
    return mock_client

class TestQAPair:
    def test_qa_pair_creation(self):
        chunk = Chunk(content="Test content", source="Test source", idx="01/01")
        qa_pair = QAPair(
            instruction="Test instruction",
            input="Test input",
            output="Test output",
            source=chunk
        )
        assert qa_pair.instruction == "Test instruction"
        assert qa_pair.input == "Test input"
        assert qa_pair.output == "Test output"
        assert qa_pair.source == chunk

    def test_qa_pair_to_dict(self):
        chunk = Chunk(content="Test content", source="Test source", idx="01/01")
        qa_pair = QAPair(
            instruction="Test instruction",
            input="Test input",
            output="Test output",
            source=chunk
        )
        result = qa_pair.to_dict()
        assert isinstance(result, dict)
        assert result["instruction"] == "Test instruction"
        assert result["input"] == "Test input"
        assert result["output"] == "Test output"
        assert "source" in result

class TestChunkStrategy:
    def test_recursive_chunk_strategy(self):
        strategy = RecursiveChunkStrategy(chunk_size=100, chunk_overlap=20)
        chunks = strategy.split("test.txt", SAMPLE_TEXT)
        
        assert isinstance(chunks, list)
        assert all(isinstance(chunk, Chunk) for chunk in chunks)
        assert all(len(chunk.content) <= 100 for chunk in chunks)

class TestMarkItDownConverter:
    def test_converter_with_text_file(self, sample_text_file):
        converter = MarkItDownConverter()
        result = converter.convert(sample_text_file)
        
        assert isinstance(result, str)
        assert "Test Document" in result
        assert "Section 1" in result
        assert "Section 2" in result

class TestAlpacaGen:
    @pytest.mark.asyncio
    async def test_generate_single_file(self, sample_text_file, mock_openai_client, tmp_path):
        output_file = tmp_path / "output.jsonl"
        
        ag = AlpacaGen(
            llm_client='azure',
            api_key='test-key',
            base_url='test-url'
        )
        
        chunks = ag.get_chunks(input_path=sample_text_file)
        dataset = ag.get_datasets(
            chunks=chunks,
            language='en',
            entries_per_chunk=1,
            output_path=output_file
        )
        
        assert isinstance(chunks, list)
        assert all(isinstance(chunk, Chunk) for chunk in chunks)
        assert isinstance(dataset, list)
        assert all(isinstance(entry, QAPair) for entry in dataset)
        assert output_file.exists()

    def test_invalid_language(self):
        ag = AlpacaGen(
            llm_client='azure',
            api_key='test-key',
            base_url='test-url'
        )
        
        with pytest.raises(AssertionError):
            ag.get_datasets(
                chunks=[],
                language='invalid'
            )

if __name__ == '__main__':
    pytest.main([__file__])
